	model	tiny
	codeseg
	org	100h
.386


;
;Macros definitions and constants
;
;data transfer delay constants (all in x/96400 secs (1 bit))
BattTOut	equ	14460;	time to wait for Battery to go active
InitWait	equ	26221;	1st file download delay
ReadWait	equ	430;	subsequent file	download delay
FlagTestDly	equ	52;	fails below 36

;data transfer delay constants (all in x/12050 secs (8 bits))
InitPad	equ	4916;	1st file upload	delay
Padding	equ	115;	subsequent file	upload delay

;data acquisition constants
BufSize	equ	7;	pulse buffer size in 2^x units
PulIgnore	equ	5;	# of initial pulses to ignore

;file xfer bar attributes/colors
InitAttr	equ	009b1h
ReadAttr	equ	00ab1h
WriteAttr	equ	00cb1h
CRCfail	equ	00d58h
AbortAttr	equ	08e30h;	value add's with 8000h+file type byte
SelAttr	equ	008h;	value or's  with vid mem
DeselAttr	equ	0f7h;	value and's with vid mem

;flags for RAM adaptor mode (379)
ScanDisk	equ	10h
Write	equ	20h
StopMotor	equ	80h
Ready	equ	08h;	37A

;flags for disk	drive mode (379)
_ScanDisk	equ	01h;	37A
_Write	equ	02h;	37A
_Ready	equ	08h
_MediaSet	equ	10h
_WriteEnable	equ	20h
_Batt	equ	80h

;mode control under disk drive operation
_ReadCond	equ	0fb81h;	cmp bx,imd16
_WriteCond	equ	0e990h;	nop;jmp rel16
rdtsc	equ	db 15,49;	machine code

;these constants are treated by	the string processor as instructions
SE	equ	0a0h;	end string inst.
LA	equ	0a1h;	load attribute inst.

;other global constants
BMcnt	equ	4;	0.055x sec benchmark time
FileSpec	equ	FileSizeTbl;	create alias
CurFileCnt	equ	gs:[-2]

;these macros are used for generating instructions used by the string processor.
OutStr	macro	Xaddr,Yaddr,Param
	local	TextStr
Tadr	=	(Yaddr*80+Xaddr)shl 1
	db	Tadr shr 9 and 15 or 80h,Tadr shr 1 and-0f01h
TextStr	db	Param
	endm

GotoStr	macro	Disp;	Disp must be word-aligned
Tadr	=	Disp-Start+100h
	db	Tadr shr 9 and 15 or 90h,Tadr shr 1 and-0f01h or(Tadr and 1)shl	8
	endm

;writes a ASCII	string of text,	then exits program.
PrintErr	macro	Param
	local	P1,P2
	mov	ah,3
	mov	bx,7
	int	10h
	mov	ax,1301h
	mov	cx,P2-P1
	lea	bp,P1
	int	10h
	int	20h
P1	db	Param
P2:	endm

;align macro
_align	macro	amt
mactemp	=	amt-(($-start)and(amt-1))
	if	mactemp ne amt
	rept	mactemp
	nop
	endm
	endif
	endm

;macro for when	a bit of data is to be transfered out to the parallel port.
OutBit	macro	reg
	out	dx,al
	mov	reg,BaseCnt
	_align	4
	dec	reg
	jnz	$-2
	endm

;macro for when	the flag status	has to be read in
InSts	macro
	in	al,dx
	mov	ebx,BaseCnt
	_align	4
	dec	ebx
	jnz	$-2
	endm


;
;Section 1:Opening FDS file & performing miscellanious checks
;
;make sure a 0 is at the end of	the command parameter
Start:	movzx	bx,ds:[80h]
	mov	byte ptr ds:[bx+81h],0

;scan through command parameter	for spaces
	mov	cx,bx
	mov	di,81h
	mov	al,' '
	cld
	rep
	scasb

;open file
	lea	dx,[di-1];	load ptr to filespec
	mov	FileSpec,dx
	mov	ax,3d02h;	dos f'n #
	int	21h
	jnc	short LoadHeader
	PrintErr	<'couldn''t open file.',13,10>

;load header
LoadHeader:	lea	dx,FDSstr;	load ptr to dest buf
	mov	cx,16;	byte count
	mov	bx,ax;	file handle
	mov	ah,3fh;	dos f'n #
	int	21h

;check header to see if authentic FDS file
	cmp	dword ptr FDSstr,1a534446h
	jz	short CheckDisks
	PrintErr	<'this is not an authentic FDS image.',13,10>

;check to see if there are any disks in FDS image (only 4 LSB of DiskCnt used)
CheckDisks:	movzx	cx,DiskCnt
	mov	dx,cx
	shl	cx,12
	jnz	short CalcAddr
	PrintErr	<'this FDS image contains no disks.',13,10>

;calculate buffer address
CalcAddr:	mov	ax,cs
	add	ax,(FreeMem-Start+10Fh)shr 4

;check to see if there is enough conventional memory available
	add	cx,ax
	jc	short NoMem
	cmp	ds:[2],cx
	jnc	short MemGood
NoMem:	PrintErr	<'not enough conventional memory to load FDS image.',13,10>

;assign temporary variables load loop will use
FileHandle	=	word ptr KeyHndlr+6
XferBarBase	=	word ptr KeyHndlr+8
DiskCtr	=	word ptr PrnHndlr+6
BufPtr	=	word ptr PrnHndlr+8
MemGood:	mov	XferBarBase,-4
	mov	FileHandle,bx
	mov	DiskCtr,dx
	mov	BufPtr,ax
	mov	BaseBufSeg,ax


;
;Section 1.3:Benchmarking system performance
;
;disable all IRQs except PIT
	in	al,21h;	get IRQ status
	mov	byte ptr IRQsts21+1,al
	push	ax
	mov	al,11111110b
	out	21h,al

;set up initial	PIT vector
	mov	cx,0b800h
	xor	ax,ax
	mov	es,cx;	set up video buffer segment
	mov	fs,ax;	set up zero page
	push	dword ptr fs:[8*4]
	mov	word ptr SPstate+1,sp
	push	cs
	push	offset PIT

;set up registers for first benchmark
	xor	esi,esi;	lo  loop count
	xor	edi,edi;	hi  loop count
	xor	ecx,ecx;	OUT loop count
	mov	dx,378h;	printer port
	lea	bx,OutBM
	mov	ax,120h
	pop	dword ptr fs:[8*4]
	jmp	$

;____________________________________________________________________________
;PIT handler
PIT:	out	20h,al
	dec	ah
	jz	$+3
	iret
	mov	ah,BMcnt
SPstate:	mov	sp,0
	sti
	jmp	bx

;____________________________________________________________________________
;benchmark 1: get loop iteration count (with OUTs) in time allocated
OutBM:	lea	bx,LoopBM;	set up next benchmark
	_align	8
	out	dx,al
	inc	ecx
	jnz	$-3

;loop count overflow occured
	lea	dx,OUTerr
BMerr:	pop	dword ptr fs:[8*4]
	pop	ax
	out	21h,al;	restore IRQ status
	mov	ah,9
	int	21h
	int	20h

;____________________________________________________________________________
;check to see if OUT benchmark provides minimum	acceptable results
LoopBM:	cmp	ecx,9531*BMcnt;192800*0.9*90112*BMcnt/1640625
	lea	dx,slowOUT
	jc	BMerr

;benchmark 2: get loop iteration count in time allocated
	lea	bx,BMdone
	_align	8
	inc	esi
	jnz	$-2
	inc	edi
	jnz	$-6

;OUT loop count	overflow occured
	lea	dx,LOOPerr
	jmp	BMerr

;____________________________________________________________________________
;error message strings
LOOPerr	db	'error occured in loop count benchmark.',13,10,'$'
OUTerr	db	'error occured during OUT operation benchmark.',13,10,'$'
slowOUT	db	'I/O port 378h is too slow for FDSLOADR to work reliably. Please consult software documentation''s "why the program may not work" section for more details.',13,10,'$'

;____________________________________________________________________________
;get ratios
BMdone:	pop	dword ptr fs:[8*4];restore old PIT handler
	pop	ax
	out	21h,al;	restore IRQ status
	mov	ebx,10590*BMcnt;192800*90112*BMcnt/1640625
	mov	eax,esi
	mov	edx,edi
	div	ecx
	mov	ebp,eax
	mov	eax,esi
	mov	edx,edi
	div	ebx

;subtract terms, clear edx if loop count is negative
	sub	eax,ebp
	cmc
	sbb	ebx,ebx
	and	eax,ebx
	inc	eax
	mov	BaseCnt,eax;	store result


;
;Section 1.6:Setting up visual displays
;
;reset video mode
	mov	ax,3
	int	10h

;display port frequency in top right corner of screen
	and	bx,0b900h
	xor	bx,0b109h
	mov	si,70*2
	mov	edx,ecx
	xor	eax,eax
	div	dword ptr OUTconv;convert to frequency
	call	BinToDec

;move cursor into next page
	mov	dx,3d4h
	mov	ax,0c0eh
	out	dx,ax
	mov	ax,770fh
	out	dx,ax

;test current connected hardware
	mov	dx,379h
	in	al,dx
	test	al,_Ready

;set up registers for ProcStr
	lea	si,RAMadpt;	load source string ptr
	mov	di,360;	set up vid mem ptr
	mov	ah,7;	load attributes	for text
	jz	ProcStr

;____________________________________________________________________________
;modify some code to prepare program for disk drive mode
	lea	si,DiskDrive;	prepare si for ProcStr
	mov	byte ptr ScanStop+1,not(_ScanDisk+_Write)
	mov	byte ptr ScanStart+1,_ScanDisk
	mov	byte ptr InitFlags+1,14h
	mov	byte ptr WaitInsert+1,_MediaSet
	mov	byte ptr TogTarget+1,SetXferMode-(TogTarget+2)
	mov	word ptr EndOfWrite,840fh
	mov	word ptr EndOfRead,840fh
	mov	word ptr WriteCond,_ReadCond
	mov	word ptr CS4,0cf58h
	jmp	ProcStr

;____________________________________________________________________________
;Disk drive mode related strings
DiskDrive	db	'Disk Drive hardware present. '
	OutStr	20, 8,<'Right Shift'>
	OutStr	36, 8,<'set xfer mode to program.'>
	OutStr	20, 9,<'Alt'>
	OutStr	36, 9,<'set xfer mode to read.'>
	OutStr	20,10,<'Spacebar'>
	OutStr	36,10,<'initiate another xfer.'>
	OutStr	 0,13,<'Current disk transfer mode is set to: '>
	db	LA,0,'REPROGRAM',LA,2ah,'READ'
	GotoStr	FDSloader

;RAM adaptor related messages
RAMadpt	db	'RAM adaptor hardware present.'
	OutStr	20, 8,<'Spacebar'>
	OutStr	36, 8,<'toggle Disk Set status.'>
	OutStr	 0,10,<'Flag status'>
	OutStr	20,11,<LA,12,'writable disk'>
	OutStr	20,12,<LA,14,'disk in drive'>
	OutStr	20,13,<LA,11,'battery good'>

;message strings shared between	modes
FDSloader:	OutStr	 0, 0,<LA,7,'FDS loader. Copyright 2002 Brad Taylor.'>
	OutStr	 0, 4,<'Hotkeys'>
	OutStr	20, 5,<'ESC'>
	OutStr	20, 6,<'F4'>
	OutStr	20, 7,<'0..9'>
	OutStr	36, 5,<'exit program.'>
	OutStr	36, 6,<'exit program and update source *.FDS file.'>
	OutStr	36, 7,<'change disk side.',SE>


;
;this procedure	is a string processor. While it	primarially moves ASCII code
;to the video memory, it also treats certain ASCII codes (80-A1) as
;instructions. The instructions	allow modifications to register	values used
;in the loop (for example, destination address). The whole idea	behind using
;a string processor is to make writing out strings to random places in video
;memory as easy	as inserting a 2-byte instruction anywhere into	the string.
;This also saves alot more memory than doing the equivelant operations in an
;x86 code sequence. Granted, this method is obviously slower, but to be fair,
;the speed bottleneck isn't in the processing of the instructions, it's in
;the writing to	video memory.
Temp2	=	$
	mov	bl,[si]
	mov	bh,al
	shl	bx,4
	sbb	cx,cx
	shr	bx,3
	sub	si,bx
	sub	di,bx
	or	si,cx
	and	di,cx
	stc
	adc	si,bx
	add	di,bx
	db	3ch
Temp	=	$
	stosw
ProcStr:	lodsb
	cmp	al,22h
	jno	Temp
	test	al,00100000b
	jz	Temp2
	mov	ah,[si]
	inc	si
	test	al,1
	jnz	ProcStr

;____________________________________________________________________________
;allocate data here for data acquisiton algorithm
;measure the distance between here and Upload, and compare it to BufSize.
;If BufSize is bigger, we back up the top of the stack apropriately.
;If Bufsize is smaller, we define the top of the stack & start of the buffer
; to be here.
NextInst	=	-4 and $-Start+103h
DifBytes	=	-4 and Upload-Start+100h-NextInst
BufBytes	=	8 shl BufSize
InitSP	=	NextInst-(BufBytes-DifBytes and	BufBytes gt DifBytes)

;define constants for keeping track of buffers
ClkBuf	equ	InitSP
PulBuf	equ	InitSP+4 shl BufSize

;initialize a new stack
SetStack:	mov	sp,InitSP and 01ffffh


;
;Section 2:Loading disk images into memory	& preparing CRC	table
;
;load FDS disk into memory
DiskLoop	=	$
	sub	dx,dx
	mov	cx,65500
	mov	ah,3fh
	mov	bx,ss:FileHandle
	mov	ds,ss:BufPtr
	add	ss:BufPtr,1000h
	mov	ds:[0],dl;	BlkLength=0 if read fails
	int	21h

;initialize some registers for loop
	mov	di,-4;	CRC destination	counter
	xor	si,si;	file pointer

;fetch file size for current file
LengthCheck	=	$
	mov	bx,-2
	mov	ax,[si-3]
	movzx	bp,[si]
	inc	ax
	shl	bp,1
	jz	short DiskCheck
	mov	ss:FileSizeTbl[4*2],ax
	sub	bx,FileSizeTbl[bp]

;update current	disk transfer progress bar
	mov	bp,ss:XferBarBase
	sub	bp,di
	sbb	cx,cx
	and	cx,(DeselAttr shl 8+0ffh and InitAttr)-InitAttr
	add	cx,InitAttr
	mov	es:[bp+16*160],cx

;calculate CRC for current file
;note that in this code body, I've used the addressing combination [ebp+ebp],
;instead of [ebp*2]. This is done because believe it or not, when ebp is used
;as an index register, it's default segment selector register is DS. This is
;also why my initial attempts failed. Even with	a "SS:" prefix,	the assembler
;*thinks* that the processor is	going to use SS	by default, and	therefore
;doesn't assemble the segment override prefix byte. Obviously, this is a
;mistake on the	assembler's behalf, but I wonder how many people know about
;this little quirk, since this specific behaviour is not well documented.
	sub	si,bx
	mov	ax,8000h
	mov	cx,[si-2]
	mov	word ptr[si-2],0
CRCloop	=	$
	movzx	ebp,al
	mov	al,ah
	mov	ah,[bx+si]
	xor	ax,CRCtbl[ebp+ebp]
	inc	bx
	jnz	CRCloop
	mov	[si-2],cx
	sub	si,2
	mov	[di],ax;	save CRC
	sub	di,2
	jmp	LengthCheck

;____________________________________________________________________________
;record number of files found to top of CRC table
DiskCheck:	mov	ds:[-2],di

;increment disk	transfer progress bar base address register
	add	ss:XferBarBase,160

;decrement disk	ctr
	dec	ss:DiskCtr
	jnz	DiskLoop


;
;Section 3:Close file, hook interrupts & enable media set
;
;set up segments
	mov	dx,cs
	mov	ds,dx
	mov	gs,BaseBufSeg

;close file
	mov	bx,FileHandle
	mov	ah,3eh
	int	21h

;intercept (and	save old) keyboard, printer, and division handlers
	shl	edx,16
	mov	ecx,edx
	mov	edi,edx
	mov	ebx,fs:[0fh*4]
	mov	eax,fs:[09h*4]
	mov	esi,fs:[00h*4]
	lea	dx,GetPul
	lea	cx,Key
	lea	di,Div0
	mov	dword ptr PrnHndlr+6,ebx
	mov	dword ptr KeyHndlr+6,eax
	mov	dword ptr DivHndlr+6,esi
	mov	fs:[0fh*4],edx
	mov	fs:[09h*4],ecx
	mov	fs:[00h*4],edi

;set up disk transfer progress bar base address	register
	mov	bp,16*160-4

;enable disk set, write enable,	& IRQ 7 interrupts
	mov	dx,37ah
InitFlags:	mov	al,17h;	enable all signals except ready
	out	dx,al
	dec	dx;	set port ptr @ 379h
	jmp	Poll


;
;Section 5:Upload FDS disk	data to RAM adaptor
;
;get block size
Upload:	movzx	bx,gs:[si]
	shl	bx,1
EndOfRead:	cmp	bx,EndScan-$-4;	neutral code filler
	mov	bx,FileSizeTbl[bx]

;replace words before & after block w/BSM & CRC	respectively.
	mov	ax,gs:[di];	CRC for current	block
	push	word ptr gs:[bx+si];save word of next block
	mov	word ptr gs:[bx+si],ax;r'place it w/ CRC
	push	word ptr gs:[si-2];save word of	last block
	mov	byte ptr gs:[si-1],80h;r'place it w/ BSM
	push	si;	save location of block start
	push	bp

;calculate loop	pointers
	lea	si,[bx+si+2];	point to end of	block
	_align	4
	add	bx,4
	add	bx,cx;	add the GAP period amount
	sub	cx,bx;	exchange registers
	neg	bx

;Send data to RAM adaptor
temp	=	$
	cmp	cx,bx
	sbb	al,al
	and	al,gs:[si+bx]
	not	al
	rept	8
	OutBit	ebp
	xor	al,1
	OutBit	ebp
	shr	al,1
	endm
	inc	bx
	jnz	temp

;restore data from previous replacement
	pop	bp
	pop	bx
	sub	si,2
	sub	bp,di
	pop	word ptr gs:[bx-2]
	pop	word ptr gs:[si]
	mov	word ptr es:[bp],ReadAttr;update file xfer bar
	add	bp,di
	sub	edi,10002h;	incement CRC ptr/set CRC flag

;____________________________________________________________________________
;wait before testing RAM adaptor's flags
ContScan:	mov	cx,FlagTestDly shl 1
temp	=	$
	OutBit	ebx
	dec	cx
	jnz	temp
	inc	dx;	379h

;load delay values for subsequent files on the disk
	mov	bx,ReadWait shl	1
	mov	cx,Padding-2

;load filesize for potential file type 4
StartScan:	mov	ax,gs:[si-3]
	inc	ax
	mov	FileSizeTbl[4*2],ax

;test ScanDisk condition flag
	in	al,dx
	dec	dx;	378h
	test	al,ScanDisk+StopMotor
	jnz	EndScan

;test Write condition
	test	al,Write
WriteCond:	jnz	Upload


;
;Section 6:Pulse interrupt	handlers (for downloading data)
;
;wait before downloading any data (to guarantee	a stable download stream)
temp	=	$
	OutBit	ecx
	dec	bx
	jnz	temp

;save register values.
	push	dx
	push	di
	push	bp

;load # of initial pulses to ignore into bx. This is done to fill up the
;buffer, since it will initially contain invalid data.
;The high word part of ebx is also being used as a counter to keep track of
;the amount of bytes to read in	before terminating the download. This is
;initially set to 0, which allows us to obtain the file type byte (the first
;byte in the file) which tells us the size of the file. Of course, the file
;size could probably be assumed, since the order the quantity of the file type
;bytes follows is usually 1,2,3,4,3,4,3,4..., but this would have made the
;code sloppier.
	mov	ebx,PulIgnore+1	shl BufSize and	0ffffh

;____________________________________________________________________________
;force pulse count register =2 for the first few unreliable interrupt pulses
GetInit:	sub	bx,1;	advance pulse ignore count
	sbb	ecx,ecx
	xor	edx,2
	sub	bx,cx;	fix up count when expired
	and	edx,ecx;	0 out or leave intact
	xor	edx,2;	load 2, or fix up from before

;test for block	start mark (should be 3 or greater)
	cmp	dx,3;	test for dx<3
	cmc
	sbb	cx,cx
	and	cx,0efh;	load 0efh if greater or equal
	sti;		enable further interrupts
	jmp	$;	wait for next interrupt

;____________________________________________________________________________
;this code calculates the # of pulses that have	elapsed.
GetPul:	sub	edx,PulBuf[di];	calculate new AbsPulCnt
	add	di,4;	increment buffer pointer
	and	di,4 shl BufSize-4;force ptr to	stay in range
	add	ebp,ClkBuf[di];	calculate BufClkCnt
	add	edx,PulBuf[di];	calculate BufPulCnt
	sub	ClkBuf[di],ebp;	store new AbsClkCnt
	sub	PulBuf[di],edx;	store AbsPulCnt
	div	ebp;	BufPulCnt/BufClkCnt
Div0:	mov	sp,InitSP-8;	deallocate interrupt stack data
	mov	ebp,eax;	load ebp with Recipricol
	mov	al,20h;	EOI command
	out	20h,al;	out to PIC
	rdtsc
	mov	edx,ebp;	load edx with Recipricol
	mov	ebp,eax;	load ebp with AbsClkCnt
	add	eax,ClkBuf[di];	calculate RelClkCnt
	mul	edx;	RelClkCnt*Recipricol
	shr	eax,31;	get 0.5 place bit
	add	edx,eax;	add it to edx
	jcxz	short GetInit

;____________________________________________________________________________
;this code assembles the data (bytes) being sent via pulses
	add	cl,dl;	add RelPulCnt to PulseAcc
	jc	short WriteByte;branch if byte is complete
WriteRtn:	ror	cl,1;	get bit to write into carry
	setc	al;	put carry into al
	rol	al,cl;	left shift al based on cl
	or	ch,al;	set apropriate bit
	mov	al,0;	zero out ax for	pulse calculations
	rol	cl,1;	restore cl
	sti;		enable further interrupts
	jmp	$;	wait for next interrupt

;____________________________________________________________________________
;when a byte is	ready to be written, this code is executed.
WriteByte:	mov	al,bl;	load al with LSB of CRC
	mov	bl,bh;	move MSB of CRC	into LSB
	mov	gs:[si],ch;	write out complete byte
	mov	bh,ch;	load MSB of CRC	with new byte
	or	cl,0f0h;	set msb of cl
	inc	si;	advance acquired data ptr
	xor	bx,CRCtbl[eax*2];CRC lookup
	mov	ch,0;	zero out byte accumulator
	sub	ebx,10000h;	decrement file byte counter
	jnc	WriteRtn;	return if write	is not done

;____________________________________________________________________________
;check to see if write is done or not
	xor	ecx,80000000h
	js	short DoneWrite

;____________________________________________________________________________
;load size for current file being written
	mov	sp,InitSP-6;	prepare stack for push
	mov	al,bh;	load al w/file type byte
	shl	ebx,16;	temporarially store bx
	mov	bx,FileSizeTbl[eax*2];load file	size
	dec	al;	types 1..4 becomes 0..3
	push	word ptr gs:[bx+si-1];push word	of next file
	rol	ebx,16;	load file count	into ebx (high word part)

;this code checks to see if the	file type byte is valid or not.	In the case
;of file type #4 (data), to pass the test, the previous file written had to
;have passed the CRC.
	bt	edi,31;	sign bit of edi	contains CRC status
	sbb	al,3;	make test
	mov	bl,80h;	move block start mark into CRC
	jc	WriteRtn;	continue if in range

;____________________________________________________________________________
;this code aborts the scan & marks the error on	the file xfer bar
;restore registers
	pop	cx;	deallocate previous push
	pop	bp
	xchg	bl,bh
	pop	di
	pop	dx

;update file transfer progress bar
	add	bx,AbortAttr
	sub	bp,di
	mov	byte ptr gs:[si-1],0;zero out invalid file's 1st byte
	mov	es:[bp],bx
	add	bp,di
	sub	di,2
	jmp	short AbortXfer

;____________________________________________________________________________
;this code wraps up the download & records the CRC status of the acquired data
;load files on disk
DoneWrite:	sub	si,2;	correct disk ptr pos'n
	cmp	bx,1;	set carry if CRC = 0
	movzx	bx,gs:[39h];	load # of files	on disk
	mov	ax,gs:[si];	load CRC value
	sbb	edi,edi;	edi = carry

;restore registers
	pop	word ptr gs:[si];replace old word
	and	di,WriteAttr-CRCfail
	pop	bp
	lea	cx,[di+CRCfail]
	pop	di
	pop	dx
	lea	bx,[ebx*4+edi+4]

;move CRC value	into table
	mov	gs:[di],ax
	sub	bp,di

;update file transfer progress bar
	mov	word ptr es:[bp],cx
	add	bp,di
	sub	di,2

;check to see if last file
	shr	bx,15
EndOfWrite:	nop
	jmp	ContScan;	only jumps on 0	in DD mode


;
;Section 4:Wait for RAM adaptor to request	data
;
;0 out last byte on disk if bx<>0
EndScan:	dec	bx
	and	gs:[si],bl

;disable Ready
AbortXfer:	mov	dl,7ah
	in	al,dx
ScanStop:	and	al,not Ready
	out	dx,al
	dec	dx;	379h

;compare old & new file counts;	erase file xfer	bar marks if new count is less
	mov	ax,di
	sub	bp,di
	sub	di,CurFileCnt
	jbe	short SkipErase
Temp	=	$
	sub	di,2
	mov	word ptr es:[bp+di],0
	ja	Temp

;record new file count
SkipErase:	mov	CurFileCnt,ax
	add	bp,ax

;restore old status of IRQ's
IRQsts21:	mov	al,0
	out	021h,al
	sti

;wait for scandisk to go inactive (or disk to be ejected in DD mode)
Temp	=	$
	in	al,dx
	test	al,ScanDisk
	jz	Temp

;hidden instructions to execute	if returning from interrupt
	dw	3d66h;	cmp eax,imd32
RepScan:	mov	sp,InitSP
	sti

;wait for scandisk to go active	(or _MediaSet &	_WriteEnable)
Poll:	mov	cx,BattTOut shl	1
Temp	=	$
	InSts
WaitInsert:	and	al,ScanDisk+StopMotor
	cmp	al,1
	sbb	cx,0
	jnc	Temp

;disable all IRQ's except 7
	cli
	mov	al,01111111b
	out	021h,al

;dehighlight file transfer bar (initiate buf ptr (si) to 0 when	done)
	mov	si,-4
	sub	si,CurFileCnt
	jz	short InitRegs
Temp	=	$
	and	byte ptr es:[bp+si+3],DeselAttr
	sub	si,2
	jnz	Temp

;initialize file/CRC ptr (di) &	"previous CRC was good" flag (bit 31 of edi)
InitRegs:	mov	edi,10000h-4;

;enable Ready (or _Scandisk)
	inc	dx;	37Ah
	in	al,dx
ScanStart:	or	al,Ready
	out	dx,al
	dec	dx;	379h

;in disk drive mode, this loop waits for _Ready	to go active, and also tests
;the battery condition. there is a timeout for the battery to go active.
;in RAM adaptor	mode, this loop	falls through.
	mov	cx,BattTOut shl	1
Temp	=	$
	InSts
	shl	al,1;	_Batt flag
	sbb	cx,0;	advance counter	if inactive
	rol	al,4;	_Ready flag
	cmc
	ja	Temp;	loop if Z=0 and	C=0

;initalize registers for initial delay values
	mov	bx,InitWait shl	1
	mov	cx,InitPad-2
	jmp	StartScan


;
;Keyboard interrupt handler
;
;exit interrupt	if repeat scan key pressed
SetXferMode:	shr	al,1
	jc	RepScan

;modify code to	accomidate requested transfer method
	shr	al,1
	sbb	ax,ax
	sbb	bx,bx
	sbb	cx,cx
	and	ax,_ReadCond	-_WriteCond
	and	bl,_ScanDisk	-(_ScanDisk+_Write)
	and	cl,2ah	-4ch
	and	bh,_MediaSet	-(_MediaSet+_WriteEnable)
	add	ax,_WriteCond
	add	bl,_ScanDisk+_Write
	add	cl,4ch
	add	bh,_MediaSet+_WriteEnable
	mov	word ptr WriteCond,ax
	mov	byte ptr ScanStart+1,bl
	mov	byte ptr WaitInsert+1,bh

;update transfer method display
	mov	bx,18h
Temp	=	$
	cmp	bx,12h
	sbb	ax,ax
	xor	al,ch
	and	al,cl
	mov	es:[bx+(13*160+77)],al
	sub	bx,2
	jnc	Temp

;exit interrupt
	pop	ax
	iret

;____________________________________________________________________________
;keyboard interrupt handler
Key:	push	ax;	save ax
	mov	al,20h;	EOI cmd
	out	20h,al;	out to PIC
	in	al,60h;	read keyboard scancode in
	sub	al,2;	check for escape scancode (1)
	jc	ExitPgm
	cmp	al,DiskCnt;	check for '0'..'9'
	jc	short ChangeSide
	sub	al,34h
	cmp	al,8;	check for status toggle keys
	jz	Save
TogTarget:	jc	short Toggle
	pop	ax;	restore ax
	iret;		exit interrupt if all tests fail

;____________________________________________________________________________
;toggle status bits
Toggle:	mov	ah,al
	push	dx
	mov	dx,37ah
	in	al,dx
	xor	al,ah
	out	dx,al
	jmp	short UDdisplay

;____________________________________________________________________________
;deselect old progress bar
ChangeSide:	mov	si,-4
	sub	si,CurFileCnt
	jz	short CS2
Temp	=	$
	and	byte ptr es:[bp+si+3],DeselAttr
	sub	si,2
	jnz	Temp

;calculate new disk transfer bar base address
CS2:	movzx	bx,al
	shl	ax,12
	imul	bx,160
	lea	bp,[bx-4+16*160]

;load appropriate BufSeg into gs
	add	ax,BaseBufSeg
	mov	gs,ax

;highlight new progress bar
	mov	si,-4
	sub	si,CurFileCnt
	jz	short CS4
Temp	=	$
	or	byte ptr es:[bp+si+3],SelAttr
	sub	si,2
	jnz	Temp

;disable disk set & write enable flags
CS4:	push	dx
	mov	dx,37ah
	in	al,dx
	and	al,-4
	out	dx,al

;update flag status display
UDdisplay:	shr	al,1
	sbb	ah,ah
	shr	al,1
	sbb	dl,dl
	shr	al,1
	sbb	dh,dh
	and	ah,12
	and	dl,14
	and	dh,11
	mov	bx,24
Temp	=	$
	mov	es:[bx+11*160+41],ah
	mov	es:[bx+12*160+41],dl
	mov	es:[bx+13*160+41],dh
	sub	bx,2
	jnc	Temp
	pop	dx

;exit interrupt
	pop	ax
	iret

;____________________________________________________________________________
;print _CantSave string
CantSave:	mov	bp,cs
	mov	ax,1301h
	mov	bx,7
	mov	cx,Save-_CantSave
	mov	dx,1800h
	mov	es,bp
	lea	bp,_CantSave
	int	10h
	jmp	short DisFlags
_CantSave	db	'Couldn''t update FDS image source file.',13,10

;____________________________________________________________________________
;open file again
Save:	mov	dx,FileSpec
	mov	ax,3d02h
	int	21h
	jc	CantSave

;offset file ptr to start of first disk image
	mov	bx,ax
	mov	dx,10h
	xor	cx,cx
	mov	ax,4200h
	int	21h

;initalize loop	variables
	mov	si,BaseBufSeg
	movzx	di,DiskCnt
	xor	dx,dx
	mov	cx,65500

;save loop- 1 iteration performaed for every disk there is
Temp	=	$
	mov	ds,si
	add	si,1000h
	mov	ah,40h
	int	21h;	write to file
	dec	di
	jnz	Temp

;close file
	mov	ah,3eh
	int	21h

;move cursor to	bottom of screen
ExitPgm:	mov	ah,2
	mov	bh,0
	mov	dx,1800h
	int	10h

;turn off status bits
DisFlags:	xor	ax,ax
	mov	dx,37ah
	out	dx,al

;restore interrupt vectors
PrnHndlr:	mov	dword ptr fs:[0fh*4],0
KeyHndlr:	mov	dword ptr fs:[09h*4],0
DivHndlr:	mov	dword ptr fs:[00h*4],0

;terminate program
	int	20h


;
;binary to decimal ASCII converter
;
;this routine accomplishes binary to decimal conversion sucessfully without
;using any division. first, the	number is multiplied by the recipricol of
;10^9. This converts the number	into a 3.29-fixed point one. in	the loop,
;each decimal digit is extracted by multiplying	the fixed-point	number by 10,
;which is easily done via binary shifts & adds.


;parameters
;
;eax	number to display
;bl	offset (controls digit alignment)
;bh	text attribute
;es:si	cursor position

BinToDec:	mul	dword ptr Recipricol
	add	eax,07079cec1h
	adc	edx,0
	mov	edi,edx
	and	edx,1fffffffh

;double-precision-multiply edx:eax by 5
;this is only required once to guarantee accuracy
	mov	ebp,edx
	mov	ecx,eax
	shld	edx,eax,2
	shl	eax,    2
	add	eax,ecx
	adc	edx,ebp

;initialize loop registers
	shr	edi,1
	mov	ch,bh;	attribute
	mov	bh,9;	loop count
	mov	ah,0;	always 0
	sub	bp,bp;	cursor increment condition
	mov	cl,-1;	ripple blanking	register

;conditional digit display loop	(9 iterations)
temp	=	$
	shr	edi,28
	setnz	al
	or	bp,ax;	=1 if non-zero digit found
	dec	al
	or	al,'0'
	and	cl,al
	cmp	bh,bl
	setbe	al
	add	di,cx
	or	bp,ax
	mov	es:[si],di
	mov	edi,edx
	add	si,bp
	and	edx,0fffffffh
	add	si,bp
	add	edx,edx
	dec	bh
	lea	edx,[edx*4+edx]
	jnz	temp

;display last digit unconditionally
	mov	cl,'0'
	shr	edi,28
	add	di,cx
	mov	es:[si],di
	ret


;
;
;
;program variables and tables
	_align	4
OUTconv	dd	235902838*BMcnt;2^32(90112/1640625)
Recipricol	dd	2305843009;	2^61/1000000000
BaseCnt	dd	?;	used for port delay control

;precalculate CRC lookup table
CRCtbl	equ	word ptr $
Ctr	=	0
	rept	256
CRC	=	Ctr
Ctr	=	Ctr+1
	rept	8
CRC	=	CRC shr 1 xor(CRC and 1)ne 0 and 8408h
	endm
	dw	CRC
	endm

FileSizeTbl	dw	?,56,2,16;	file type size LUT
FDSstr	dw	?
BaseBufSeg	dw	?;	segment address	of free mem
DiskCnt	db	?;	disk count of loaded *.FDS file

FreeMem:	ends
	end	Start
